﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Reflection;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Execution;
using OpcLabs.BaseLib.Extensions;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Reflection.Extensions;
using OpcLabs.BaseLib.Utilities;

namespace OpcLabs.UnitTestLib.Testers
{
    public class ShouldSerializeTester : Tester
    {
        private class Worker : MarshalByRefObject
        {
            public void Test(
                [NotNull] Type typeUnderTest,
                [NotNull] MemberInfo dataMember,
                [NotNull] MethodInfo shouldSerializeMethod)
            {
                object instance = null;
                if (dataMember.IsStatic())
                    AssertUtilities.WriteLine("Data member is static.");
                else
                {
                    AssertUtilities.WriteLine("Creating instance...");
                    instance = Activator.CreateInstance(typeUnderTest, nonPublic: true);
                }
                AssertUtilities.WriteLine("Calling ShouldSerializeXXXX...");
                object value = shouldSerializeMethod.Invoke(instance, null);
                Assert.IsNotNull(value);
                bool result = (bool)value;
                Assert.AreEqual(false, result, $"{shouldSerializeMethod} has returned True.");
                AssertUtilities.WriteLine("Done.");
            }
        }

        public ShouldSerializeTester() 
        {
            TestCategories.Add("DataMembers");
        }

        [NotNull]
        static private TestMethod CreateTestMethod(
            [NotNull] Type typeUnderTest,
            [NotNull] MemberInfo dataMember,
            [NotNull] MethodInfo shouldSerializeMethod)
        {
            return new DelegatedTestMethod(
                string.Format(CultureInfo.InvariantCulture, "AfterDefaultCtor_{0}", dataMember.Name), 
                () =>
                    {
                        if (dataMember.IsStatic())
                        {
#if (NETFRAMEWORK)
                            using (var isolatedApp = new IsolatedApp<Worker>())
                                isolatedApp.Value.Test(typeUnderTest, dataMember, shouldSerializeMethod);
#else
                            AssertUtilities.Inconclusive("Testing static members is not supported under .NET Standard.");
#endif
                        }
                        else
                            new Worker().Test(typeUnderTest, dataMember, shouldSerializeMethod);
                    });
        }

        protected override void Initialize(Type typeUnderTest)
        {
            base.Initialize(typeUnderTest);

            bool hasDefaultConstructor = typeUnderTest.HasDefaultConstructor();

            IEnumerable<MemberInfo> dataMembers = TypeUtilities.HighestLevelMembers(typeUnderTest, typeUnderTest.GetDataMembers());
            foreach (MemberInfo dataMember in dataMembers)
            {
                Debug.Assert(dataMember != null);
                if (!dataMember.IsStatic() && !hasDefaultConstructor)
                    continue;
                if (!ShouldTestDataMember(dataMember))
                    continue;

                MethodInfo shouldSerializeMethod = typeUnderTest.GetMethod("ShouldSerialize" + dataMember.Name,
                    (dataMember.IsStatic() ? BindingFlags.Static : BindingFlags.Instance) | BindingFlags.Public | BindingFlags.NonPublic,
                    null, Type.EmptyTypes, null);
                if (shouldSerializeMethod == null)
                    continue;

                AddTestMethod(CreateTestMethod(typeUnderTest, dataMember, shouldSerializeMethod));
            }
        }

        protected virtual bool ShouldTestDataMember([NotNull] MemberInfo dataMember)
        {
            Debug.Assert(dataMember != null);
            // ReSharper disable once ConvertIfStatementToReturnStatement
            if (!dataMember.IsSettable())
                return false;
            return true;
        }
    }
}
