﻿
using System;
using System.Diagnostics;
using System.Net.Sockets;
#if (NETFRAMEWORK)
using System.ServiceModel;
#endif
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.Navigation;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.UA
{
    partial class EasyUAClientTest
    {
        /// <summary>
        /// Tests a Read method for an attribute that has Bad quality in StatusCode.
        /// </summary>
        /// <remarks>
        /// The tests "BadQualityXXXX" and "UncertainQualityXXXX" were written to verify the problem and the fix where Read or 
        /// ReadMultiple would return an exception instead of valid UAAttributeData.
        /// </remarks>
        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void BadQualityReadTest([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_UInt16Value;
            var writeAttributeData = new UAAttributeData(123, UAStatusCode.FromUASeverity(UASeverity.BadOrFailure));

            target.Write(endpointDescriptor, nodeId, writeAttributeData, TypeCode.UInt16);

            TestSection.Act();
            UAAttributeData attributeData = target.Read(endpointDescriptor, nodeId);

            TestSection.Assert();

            Assert.IsNotNull(attributeData);
            Assert.AreEqual(writeAttributeData.StatusCode, attributeData.StatusCode);
            // With UASeverity.BadOrFailure, equality with the value written cannot be asserted
        }

        /// <summary>
        /// Tests a ReadMultiple method for an attribute that has Bad quality in StatusCode.
        /// </summary>
        /// <remarks>
        /// The tests "BadQualityXXXX" and "UncertainQualityXXXX" were written to verify the problem and the fix where Read or 
        /// ReadMultiple would return an exception instead of valid UAAttributeData.
        /// </remarks>
        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void BadQualityReadMultipleTest([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_UInt16Value;
            var writeAttributeData = new UAAttributeData(123, UAStatusCode.FromUASeverity(UASeverity.BadOrFailure));
            target.Write(endpointDescriptor, nodeId, writeAttributeData, TypeCode.UInt16);

            TestSection.Act();
            UAAttributeDataResult[] readAttributeDataResults =
                target.ReadMultiple(new[] { new UAReadArguments(endpointDescriptor, nodeId) });

            TestSection.Assert();

            Assert.IsNotNull(readAttributeDataResults);
            Assert.AreEqual(1, readAttributeDataResults.Length);
            UAAttributeDataResult attributeDataResult = readAttributeDataResults[0];
            Assert.IsNotNull(attributeDataResult);
            Assert.IsNull(attributeDataResult.State);
            Assert.IsNull(attributeDataResult.Exception);
            UAAttributeData attributeData = attributeDataResult.AttributeData;
            Assert.IsNotNull(attributeData);
            Assert.AreEqual(writeAttributeData.StatusCode, attributeData.StatusCode);
            // With UASeverity.BadOrFailure, equality with the value written cannot be asserted
        }

        /// <summary>
        /// Tests a ReadValue method for an attribute that has Bad quality in StatusCode.
        /// </summary>
        /// <remarks>
        /// The tests "BadQualityXXXX" and "UncertainQualityXXXX" were written to verify the problem and the fix where Read or 
        /// ReadMultiple would return an exception instead of valid UAAttributeData.
        /// </remarks>
        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void BadQualityReadValueTest([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_UInt16Value;
            UAStatusCode statusCode = UAStatusCode.FromUASeverity(UASeverity.BadOrFailure);
            var writeAttributeData = new UAAttributeData(123, statusCode);
            target.Write(endpointDescriptor, nodeId, writeAttributeData, TypeCode.UInt16);

            try
            {
                object value;
                TestSection.Act();
                try
                {
                    value = target.ReadValue(endpointDescriptor, nodeId);
                }
                finally
                {
                    TestSection.Assert();
                }
                AssertUtilities.WriteLine(value);
            }
            catch (UAException uaException)
            {
                Exception innerException = uaException.InnerException;
                Assert.IsNotNull(innerException);
                var statusCodeException = innerException as UAStatusCodeException;
                Assert.IsNotNull(statusCodeException);
                Assert.AreEqual(writeAttributeData.StatusCode, statusCodeException.StatusCode);
                return;
            }
            Assert.Fail("UAException not thrown.");
        }

        [TestCategory("OpcUA")]
        [TestCategory("Live")]  // IMPROVE: not really live; move elsewhere
        [TestMethod]
        public void ReadMultipleValuesTest_ReadArguments_None()
        {
            TestSection.Arrange();
            UAReadArguments[] readArgumentsArray = new UAReadArguments[0];
            var target = new EasyUAClient();

            TestSection.Act();
            ValueResult[] valueResultArray = target.ReadMultipleValues(readArgumentsArray);

            TestSection.Assert();
            Assert.IsNotNull(valueResultArray);
            Assert.AreEqual(0, valueResultArray.Length);
        }

        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ReadMultipleValuesTest_ReadArguments_One()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor = UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_SByteValue;
            UAReadArguments[] readArgumentsArray = { new UAReadArguments(endpointDescriptor, nodeId) };

            var target = new EasyUAClient();

            TestSection.Act();
            ValueResult[] valueResultArray = target.ReadMultipleValues(readArgumentsArray);

            TestSection.Assert();

            Assert.IsNotNull(valueResultArray);
            Assert.AreEqual(1, valueResultArray.Length);

            ValueResult valueResult = valueResultArray[0];
            Assert.IsNotNull(valueResult);
            AssertUtilities.WriteLine(valueResult);
            Assert.IsNull(valueResult.Exception);
            object value = valueResult.Value;
            Assert.IsNotNull(value);
            AssertUtilities.WriteLine(value);
            Assert.IsInstanceOfType(value, typeof(Int16));
        }

        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ReadMultipleValuesTest_ReadArguments_Two_IdenticalReadArguments()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor = UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_Int32Value;
            UAReadArguments readArguments = new UAReadArguments(endpointDescriptor, nodeId);
            UAReadArguments[] readArgumentsArray = { readArguments, readArguments };
            var target = new EasyUAClient();

            TestSection.Act();
            ValueResult[] valueResultArray = target.ReadMultipleValues(readArgumentsArray);

            TestSection.Assert();

            Assert.IsNotNull(valueResultArray);
            Assert.AreEqual(2, valueResultArray.Length);

            ValueResult valueResult0 = valueResultArray[0];
            Assert.IsNotNull(valueResult0);
            AssertUtilities.WriteLine(valueResult0);
            Assert.IsNull(valueResult0.Exception);
            object value0 = valueResult0.Value;
            Assert.IsNotNull(value0);
            AssertUtilities.WriteLine(value0);
            Assert.IsInstanceOfType(value0, typeof(Int32));

            ValueResult valueResult1 = valueResultArray[1];
            Assert.IsNotNull(valueResult1);
            AssertUtilities.WriteLine(valueResult1);
            Assert.IsNull(valueResult1.Exception);
            object value1 = valueResult1.Value;
            Assert.IsNotNull(value1);
            AssertUtilities.WriteLine(value1);
            Assert.IsInstanceOfType(value1, typeof(Int32));

            Assert.AreEqual(value0, value1);
        }

        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ReadMultipleValuesTest_ReadArguments_Two_IdenticalEndpointDescriptors_IdenticalNodeDescriptors()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor = UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_Int32Value;
            UANodeDescriptor nodeDescriptor = nodeId;

            UAReadArguments readArguments0 = new UAReadArguments(endpointDescriptor, nodeDescriptor);
            UAReadArguments readArguments1 = new UAReadArguments(endpointDescriptor, nodeDescriptor);
            Debug.Assert(readArguments0 == readArguments1);
            UAReadArguments[] readArgumentsArray = { readArguments0, readArguments1 };

            var target = new EasyUAClient();

            TestSection.Act();
            ValueResult[] valueResultArray = target.ReadMultipleValues(readArgumentsArray);

            TestSection.Assert();

            Assert.IsNotNull(valueResultArray);
            Assert.AreEqual(2, valueResultArray.Length);

            ValueResult valueResult0 = valueResultArray[0];
            Assert.IsNotNull(valueResult0);
            AssertUtilities.WriteLine(valueResult0);
            Assert.IsNull(valueResult0.Exception);
            object value0 = valueResult0.Value;
            Assert.IsNotNull(value0);
            AssertUtilities.WriteLine(value0);
            Assert.IsInstanceOfType(value0, typeof(Int32));

            ValueResult valueResult1 = valueResultArray[1];
            Assert.IsNotNull(valueResult1);
            AssertUtilities.WriteLine(valueResult1);
            Assert.IsNull(valueResult1.Exception);
            object value1 = valueResult1.Value;
            Assert.IsNotNull(value1);
            AssertUtilities.WriteLine(value1);
            Assert.IsInstanceOfType(value1, typeof(Int32));

            Assert.AreEqual(value0, value1);
        }

        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ReadMultipleValuesTest_ReadArguments_Two_IdenticalEndpointDescriptors_EqualNodeDescriptors()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor = UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_Int32Value;
            UANodeDescriptor nodeDescriptor0 = nodeId;
            UANodeDescriptor nodeDescriptor1 = nodeId;
            Debug.Assert(nodeDescriptor0 == nodeDescriptor1);

            UAReadArguments readArguments0 = new UAReadArguments(endpointDescriptor, nodeDescriptor0);
            UAReadArguments readArguments1 = new UAReadArguments(endpointDescriptor, nodeDescriptor1);
            Debug.Assert(readArguments0 == readArguments1);
            UAReadArguments[] readArgumentsArray = { readArguments0, readArguments1 };

            var target = new EasyUAClient();

            TestSection.Act();
            ValueResult[] valueResultArray = target.ReadMultipleValues(readArgumentsArray);

            TestSection.Assert();

            Assert.IsNotNull(valueResultArray);
            Assert.AreEqual(2, valueResultArray.Length);

            ValueResult valueResult0 = valueResultArray[0];
            Assert.IsNotNull(valueResult0);
            AssertUtilities.WriteLine(valueResult0);
            Assert.IsNull(valueResult0.Exception);
            object value0 = valueResult0.Value;
            Assert.IsNotNull(value0);
            AssertUtilities.WriteLine(value0);
            Assert.IsInstanceOfType(value0, typeof(Int32));

            ValueResult valueResult1 = valueResultArray[1];
            Assert.IsNotNull(valueResult1);
            AssertUtilities.WriteLine(valueResult1);
            Assert.IsNull(valueResult1.Exception);
            object value1 = valueResult1.Value;
            Assert.IsNotNull(value1);
            AssertUtilities.WriteLine(value1);
            Assert.IsInstanceOfType(value1, typeof(Int32));

            Assert.AreEqual(value0, value1);
        }

        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ReadMultipleValuesTest_ReadArguments_Two_IdenticalEndpointDescriptors_UnequalNodeDescriptors()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor = UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            UANodeId nodeId0 = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_Int32Value;
            UANodeId nodeId1 = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_UserScalar_Int32Value;
            UANodeDescriptor nodeDescriptor0 = nodeId0;
            UANodeDescriptor nodeDescriptor1 = nodeId1;
            Debug.Assert(nodeDescriptor0 != nodeDescriptor1);

            UAReadArguments readArguments0 = new UAReadArguments(endpointDescriptor, nodeDescriptor0);
            UAReadArguments readArguments1 = new UAReadArguments(endpointDescriptor, nodeDescriptor1);
            Debug.Assert(readArguments0 != readArguments1);
            UAReadArguments[] readArgumentsArray = { readArguments0, readArguments1 };

            var target = new EasyUAClient();

            TestSection.Act();
            ValueResult[] valueResultArray = target.ReadMultipleValues(readArgumentsArray);

            TestSection.Assert();

            Assert.IsNotNull(valueResultArray);
            Assert.AreEqual(2, valueResultArray.Length);

            ValueResult valueResult0 = valueResultArray[0];
            Assert.IsNotNull(valueResult0);
            AssertUtilities.WriteLine(valueResult0);
            Assert.IsNull(valueResult0.Exception);
            object value0 = valueResult0.Value;
            Assert.IsNotNull(value0);
            AssertUtilities.WriteLine(value0);
            Assert.IsInstanceOfType(value0, typeof(Int32));

            ValueResult valueResult1 = valueResultArray[1];
            Assert.IsNotNull(valueResult1);
            AssertUtilities.WriteLine(valueResult1);
            Assert.IsNull(valueResult1.Exception);
            object value1 = valueResult1.Value;
            Assert.IsNotNull(value1);
            AssertUtilities.WriteLine(value1);
            Assert.IsInstanceOfType(value1, typeof(Int32));
        }

        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ReadMultipleValuesTest_ReadArguments_Two_EqualEndpointDescriptors()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor0 = UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            UAEndpointDescriptor endpointDescriptor1 = UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            Debug.Assert(endpointDescriptor0 == endpointDescriptor1);

            TestEnvironment.PrecheckEndpoint(endpointDescriptor0);
            TestEnvironment.PrecheckEndpoint(endpointDescriptor1);

            UANodeId nodeId0 = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_Int32Value;
            UANodeId nodeId1 = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_UserScalar_Int32Value;
            UANodeDescriptor nodeDescriptor0 = nodeId0;
            UANodeDescriptor nodeDescriptor1 = nodeId1;
            Debug.Assert(nodeDescriptor0 != nodeDescriptor1);

            UAReadArguments readArguments0 = new UAReadArguments(endpointDescriptor0, nodeDescriptor0);
            UAReadArguments readArguments1 = new UAReadArguments(endpointDescriptor1, nodeDescriptor1);
            Debug.Assert(readArguments0 != readArguments1);
            UAReadArguments[] readArgumentsArray = { readArguments0, readArguments1 };

            var target = new EasyUAClient();

            TestSection.Act();
            ValueResult[] valueResultArray = target.ReadMultipleValues(readArgumentsArray);

            TestSection.Assert();

            Assert.IsNotNull(valueResultArray);
            Assert.AreEqual(2, valueResultArray.Length);

            ValueResult valueResult0 = valueResultArray[0];
            Assert.IsNotNull(valueResult0);
            AssertUtilities.WriteLine(valueResult0);
            Assert.IsNull(valueResult0.Exception);
            object value0 = valueResult0.Value;
            Assert.IsNotNull(value0);
            AssertUtilities.WriteLine(value0);
            Assert.IsInstanceOfType(value0, typeof(Int32));

            ValueResult valueResult1 = valueResultArray[1];
            Assert.IsNotNull(valueResult1);
            AssertUtilities.WriteLine(valueResult1);
            Assert.IsNull(valueResult1.Exception);
            object value1 = valueResult1.Value;
            Assert.IsNotNull(value1);
            AssertUtilities.WriteLine(value1);
            Assert.IsInstanceOfType(value1, typeof(Int32));
        }

        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ReadMultipleValuesTest_ReadArguments_Two_UnequalEndpointDescriptors()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor0 = UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            UAEndpointDescriptor endpointDescriptor1 = UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.Https;
            Debug.Assert(endpointDescriptor0 != endpointDescriptor1);
            
            if (!UALiveTestEnvironment.IsSupportedOptimistic(endpointDescriptor0))
                Assert.Inconclusive($"Endpoint not optimistically supported: {endpointDescriptor0}");
            if (!UALiveTestEnvironment.IsSupportedOptimistic(endpointDescriptor1))
                Assert.Inconclusive($"Endpoint not optimistically supported: {endpointDescriptor1}");

            TestEnvironment.PrecheckEndpoint(endpointDescriptor0);
            TestEnvironment.PrecheckEndpoint(endpointDescriptor1);

            UANodeId nodeId0 = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_Int32Value;
            UANodeId nodeId1 = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_UserScalar_Int32Value;
            UANodeDescriptor nodeDescriptor0 = nodeId0;
            UANodeDescriptor nodeDescriptor1 = nodeId1;
            Debug.Assert(nodeDescriptor0 != nodeDescriptor1);

            UAReadArguments readArguments0 = new UAReadArguments(endpointDescriptor0, nodeDescriptor0);
            UAReadArguments readArguments1 = new UAReadArguments(endpointDescriptor1, nodeDescriptor1);
            Debug.Assert(readArguments0 != readArguments1);
            UAReadArguments[] readArgumentsArray = { readArguments0, readArguments1 };

            var target = new EasyUAClient();

            TestSection.Act();
            ValueResult[] valueResultArray = target.ReadMultipleValues(readArgumentsArray);

            TestSection.Assert();

            Assert.IsNotNull(valueResultArray);
            Assert.AreEqual(2, valueResultArray.Length);

            ValueResult valueResult0 = valueResultArray[0];
            Assert.IsNotNull(valueResult0);
            AssertUtilities.WriteLine(valueResult0);
            Assert.IsNull(valueResult0.Exception);
            object value0 = valueResult0.Value;
            Assert.IsNotNull(value0);
            AssertUtilities.WriteLine(value0);
            Assert.IsInstanceOfType(value0, typeof(Int32));

            ValueResult valueResult1 = valueResultArray[1];
            Assert.IsNotNull(valueResult1);
            AssertUtilities.WriteLine(valueResult1);
            Assert.IsNull(valueResult1.Exception);
            object value1 = valueResult1.Value;
            Assert.IsNotNull(value1);
            AssertUtilities.WriteLine(value1);
            Assert.IsInstanceOfType(value1, typeof(Int32));
        }

        // IMPROVE: put this back
        //[DynamicData(nameof(UATestEnvironment.OpcFoundation_SampleServer_AllEndpointDescriptors), typeof(UATestEnvironment))]
        //[TestCategory("Live")]
        //[TestCategory("OpcUA")]
        //[TestMethod]
        //public void ReadTest_BadCertificateUntrusted([NotNull] UAEndpointDescriptor endpointDescriptor)
        //{
        //    TestSection.Arrange();
        //    TestEnvironment.PrecheckEndpoint(endpointDescriptor);

        //    EasyUAClient.ClearValidatedCertificates();
        //    var target = new EasyUAClient { Isolated = true };
        //    //target.IsolatedParameters.SessionParameters.CertificateAcceptancePolicy.AcceptAnyCertificate = false;
        //    //target.IsolatedParameters.SessionParameters.CertificateAcceptancePolicy.TrustedEndpointUrlStrings = new StringCollection();
        //    //target.IsolatedParameters.SessionParameters.CertificateAcceptancePolicy.AllowUserAcceptCertificate = false;
        //    //Assert.AreEqual(0, target.IsolatedParameters.SessionParameters.CertificateAcceptancePolicy.TrustedEndpointUrlStrings.Count);
        //    target.IsolatedParameters.SessionParameters.EndpointSelectionPolicy.AllowedMessageSecurityModes = UAMessageSecurityModes.SecuritySign;

        //    UANodeId nodeId = OpcFoundation_SampleServer.NodeIdExpandedTexts.Data_Static_Scalar_Int32Value;
        //    try
        //    {
        //        TestSection.Act();
        //        try
        //        {
        //            target.Read(endpointDescriptor, nodeId);
        //        }
        //        finally
        //        {
        //            TestSection.Assert();
        //        }
        //    }
        //    catch (UAException uaException)
        //    {
        //        const int badCertificateUntrusted = UACodeBits.BadCertificateUntrusted;
        //        int errorCode = uaException.ErrorCode;
        //        Assert.AreEqual(badCertificateUntrusted, ((UAStatusCode)unchecked((uint)errorCode)).CodeBits);
        //        Exception innerException = uaException.InnerException;
        //        Assert.IsNotNull(innerException);
        //        var serviceException = innerException as UAServiceException;
        //        Assert.IsNotNull(serviceException);
        //        Assert.AreEqual(badCertificateUntrusted, serviceException.ServiceResult.StatusCode.CodeBits);
        //        // IMPROVE: assert more
        //        Assert.IsNull(serviceException.InnerException);
        //        return;
        //    }
        //    Assert.Fail("UAException not thrown.");
        //}

        // This test intentionally does not use [DynamicData], so that we can use it for debugging HTTPS specifically.
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void ReadTest_Https()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor =
                UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.Https;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_Scalar_Int32Value;
            target.WriteValue(endpointDescriptor, nodeId, 893586, TypeCode.Int32);

            TestSection.Act();
            UAAttributeData attributeData = target.Read(endpointDescriptor, nodeId);

            TestSection.Assert();
            Assert.IsNotNull(attributeData);
            object value = attributeData.Value;
            Assert.IsNotNull(value);
            Assert.AreEqual(893586, value);
            // IMPROVE: assert about timestamps, status
        }

        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void ReadTest_UnknownNodeId([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Unknown;

            TestSection.Act();
            UAAttributeData attributeData = target.Read(endpointDescriptor, nodeId);

            TestSection.Assert();
            Assert.IsNotNull(attributeData);
            Assert.IsTrue(attributeData.HasBadStatus);
            Assert.IsFalse(attributeData.HasGoodStatus);
            Assert.IsFalse(attributeData.HasUncertainStatus);
            Assert.IsFalse(attributeData.HasValue);
            Assert.AreEqual(UACodeBits.BadNodeIdUnknown, attributeData.StatusCode.CodeBits);
        }

        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void ReadValueTest([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_Scalar_Int32Value;
            target.WriteValue(endpointDescriptor, nodeId, 893586, TypeCode.Int32);

            TestSection.Act();
            object value = target.ReadValue(endpointDescriptor, nodeId);

            TestSection.Assert();
            Assert.IsNotNull(value);
            Assert.AreEqual(893586, value);
        }

        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void ReadValueTest_BrowsePath([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UABrowsePath browsePath = UABrowsePath.Parse(
                UABrowsePath.Null,
                OpcFoundation_UASampleServer.BrowsePathStrings.Data_Static_UserScalar_Int32Value,
                OpcFoundation_UASampleServer.NamespaceUriStrings.Default);
            target.WriteValue(endpointDescriptor, browsePath, 893587, TypeCode.Int32);

            TestSection.Act();
            object value = target.ReadValue(endpointDescriptor, browsePath);

            TestSection.Assert();
            Assert.IsNotNull(value);
            Assert.AreEqual(893587, value);
        }

        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ReadValueTest_Dynamic([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_SByteValue;

            TestSection.Act();
            object value = target.ReadValue(endpointDescriptor, nodeId);

            TestSection.Assert();
            Assert.IsNotNull(value);
            AssertUtilities.WriteLine(value);
            Assert.IsInstanceOfType(value, typeof(Int16));
        }

        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_SynthesizableEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void ReadValueTest_EndpointSynthesis([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_Scalar_Int32Value;
            target.WriteValue(endpointDescriptor, nodeId, 893586, TypeCode.Int32);

            TestSection.Act();
            object value = target.ReadValue(endpointDescriptor, nodeId);

            TestSection.Assert();
            Assert.IsNotNull(value);
            Assert.AreEqual(893586, value);
        }

#if (NETFRAMEWORK)
        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ReadValueTest_OpcHttp()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.Http;
            endpointDescriptor.Scheme = "opc.http";
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_SByteValue;

            TestSection.Act();
            object value = target.ReadValue(endpointDescriptor, nodeId);

            TestSection.Assert();
            Assert.IsNotNull(value);
            AssertUtilities.WriteLine(value);
            Assert.IsInstanceOfType(value, typeof(Int16));
        }
#endif

        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ReadValueTest_UnknownHost_OpcTcp_Simple()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = "opc.tcp://unknown.host.BA10BDC9-C492-403A-B3C1-01838123C3E3:41211/UA/SampleServer";
            //TestEnvironment.Precheck(endpointDescriptor); // not here

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_Scalar_Int32Value;

            try
            {
                TestSection.Act();
                try
                {
                    target.ReadValue(endpointDescriptor, nodeId);
                }
                finally
                {
                    TestSection.Assert();
                }
            }
            catch (UAException exception)
            {
                AssertUtilities.WriteLine(exception);
                string message = exception.Message;
#if (NETFRAMEWORK)
                StringAssert.Contains(message, "System.Net.Sockets.SocketException");
                Exception innerException = exception.InnerException;
                Assert.IsInstanceOfType(innerException, typeof(SocketException));
#else
                StringAssert.Contains(message, "'BadNotConnected'");
                Exception innerException = exception.InnerException;
                Assert.IsInstanceOfType(innerException, typeof(UAServiceException));
#endif
            }
        }

#if (NETFRAMEWORK)
        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ReadValueTest_UnknownHost_Http_Simple()
        {
            TestSection.Arrange();
            var target = new EasyUAClient();
            UAEndpointDescriptor endpointDescriptor = "http://unknown.host.A7A22390-A9AA-428E-9980-3DDE35AD82A9:41211/UA/SampleServer";
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_Scalar_Int32Value;

            try
            {
                TestSection.Act();
                try
                {
                    target.ReadValue(endpointDescriptor, nodeId);
                }
                finally
                {
                    TestSection.Arrange();
                }
            }
            catch (UAException exception)
            {
                AssertUtilities.WriteLine(exception);
                string message = exception.Message;
                StringAssert.Contains(message, "System.ServiceModel.EndpointNotFoundException");
                Exception innerException = exception.InnerException;
                Assert.IsInstanceOfType(innerException, typeof(EndpointNotFoundException));
            }
        }
#endif

        [TestCategory("OpcUA")]
        [TestCategory("Live")]
        [TestMethod]
        public void ReadValueTest_UnknownHost_Https_Simple()
        {
            TestSection.Arrange();
            var target = new EasyUAClient();
            UAEndpointDescriptor endpointDescriptor = "https://unknown.host.A7A22390-A9AA-428E-9980-3DDE35AD82A9:41211/UA/SampleServer";
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_Scalar_Int32Value;

            try
            {
                TestSection.Act();
                try
                {
                    target.ReadValue(endpointDescriptor, nodeId);
                }
                finally
                {
                    TestSection.Arrange();
                }
            }
            catch (UAException exception)
            {
                AssertUtilities.WriteLine(exception);
                string message = exception.Message;
#if (NETFRAMEWORK)
                StringAssert.Contains(message, "'BadCommunicationError'");
#else
                StringAssert.Contains(message, "'BadUnknownResponse'");
#endif
                Exception innerException = exception.InnerException;
                Assert.IsInstanceOfType(innerException, typeof(UAServiceException));
            }
        }

        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        [ExpectedException(typeof(UAException))]
        public void ReadValueTest_UnknownNodeId_Simple([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Unknown;

            TestSection.Act();
            target.ReadValue(endpointDescriptor, nodeId);
        }

        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void ReadValueTest_UserIdentity_UserName([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptorWithUserIdentity = new UAEndpointDescriptor(endpointDescriptor);
            endpointDescriptorWithUserIdentity.UserIdentity.UserNameTokenInfo.UserName = "user";
            endpointDescriptorWithUserIdentity.UserIdentity.UserNameTokenInfo.Password = "password";
            TestEnvironment.PrecheckEndpoint(endpointDescriptorWithUserIdentity);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_Scalar_Int32Value;
            target.WriteValue(endpointDescriptorWithUserIdentity, nodeId, 893586, TypeCode.Int32);

            TestSection.Act();
            object value = target.ReadValue(endpointDescriptorWithUserIdentity, nodeId);

            TestSection.Assert();
            Assert.IsNotNull(value);
            Assert.AreEqual(893586, value);
        }

        /// <summary>
        /// Tests a Read method for an attribute that has Uncertain quality in StatusCode.
        /// </summary>
        /// <remarks>
        /// The tests "BadQualityXXXX" and "UncertainQualityXXXX" were written to verify the problem and the fix where Read or 
        /// ReadMultiple would return an exception instead of valid UAAttributeData.
        /// </remarks>
        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void UncertainQualityReadTest([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_UInt16Value;
            var writeAttributeData = new UAAttributeData(123, UAStatusCode.FromUASeverity(UASeverity.UncertainOrWarning));
            target.Write(endpointDescriptor, nodeId, writeAttributeData, TypeCode.UInt16);

            TestSection.Act();
            UAAttributeData attributeData = target.Read(endpointDescriptor, nodeId);

            TestSection.Assert();
            Assert.IsNotNull(attributeData);
            Assert.AreEqual(writeAttributeData.StatusCode, attributeData.StatusCode);
            Assert.AreEqual(writeAttributeData.Value, attributeData.Value);
        }

        /// <summary>
        /// Tests a ReadMultiple method for an attribute that has Uncertain quality in StatusCode.
        /// </summary>
        /// <remarks>
        /// The tests "BadQualityXXXX" and "UncertainQualityXXXX" were written to verify the problem and the fix where Read or 
        /// ReadMultiple would return an exception instead of valid UAAttributeData.
        /// </remarks>
        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void UncertainQualityReadMultipleTest([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_UInt16Value;
            var writeAttributeData = new UAAttributeData(123, UAStatusCode.FromUASeverity(UASeverity.UncertainOrWarning));
            target.Write(endpointDescriptor, nodeId, writeAttributeData, TypeCode.UInt16);

            TestSection.Act();
            UAAttributeDataResult[] readAttributeDataResults =
                target.ReadMultiple(new[] { new UAReadArguments(endpointDescriptor, nodeId) });

            TestSection.Assert();
            Assert.IsNotNull(readAttributeDataResults);
            Assert.AreEqual(1, readAttributeDataResults.Length);
            UAAttributeDataResult attributeDataResult = readAttributeDataResults[0];
            Assert.IsNotNull(attributeDataResult);
            Assert.IsNull(attributeDataResult.State);
            Assert.IsNull(attributeDataResult.Exception);
            UAAttributeData attributeData = attributeDataResult.AttributeData;
            Assert.IsNotNull(attributeData);
            Assert.AreEqual(writeAttributeData.StatusCode, attributeData.StatusCode);
            Assert.AreEqual(writeAttributeData.Value, attributeData.Value);
        }

        /// <summary>
        /// Tests a ReadValue method for an attribute that has Uncertain quality in StatusCode.
        /// </summary>
        /// <remarks>
        /// The tests "BadQualityXXXX" and "UncertainQualityXXXX" were written to verify the problem and the fix where Read or 
        /// ReadMultiple would return an exception instead of valid UAAttributeData.
        /// </remarks>
        [DynamicData(nameof(UALiveTestEnvironment.OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows), typeof(UALiveTestEnvironment))]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void UncertainQualityReadValueTest([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var target = new EasyUAClient();
            UANodeId nodeId = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_UInt16Value;
            var writeAttributeData = new UAAttributeData(123, UAStatusCode.FromUASeverity(UASeverity.UncertainOrWarning));
            target.Write(endpointDescriptor, nodeId, writeAttributeData, TypeCode.UInt16);

            try
            {
                object value;
                TestSection.Act();
                try
                {
                    value = target.ReadValue(endpointDescriptor, nodeId);
                }
                finally
                {
                    TestSection.Assert();
                }
                AssertUtilities.WriteLine(value);
            }
            catch (UAException uaException)
            {
                Exception innerException = uaException.InnerException;
                Assert.IsNotNull(innerException);
                var statusCodeException = innerException as UAStatusCodeException;
                Assert.IsNotNull(statusCodeException);
                Assert.AreEqual(writeAttributeData.StatusCode, statusCodeException.StatusCode);
                return;
            }
            Assert.Fail("UAException not thrown.");
        }
    }
}
