﻿
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.BaseLib.DataTypeModel;
using OpcLabs.BaseLib.IO.Extensions;
using OpcLabs.BaseLib.Utilities;
using OpcLabs.EasyOpc.UA.DataTypeModel;

namespace OpcLabs.EasyOpcTest.UA.DataTypeModel.Internal
{
    [TestClass]
    public class BuildingAutomationTest : UAOpcBinaryDataTypeSystemTest
    {
        #region Static Public Properties

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public GenericData BAConfigurationsData
        {
            get
            {
                StructuredData CreateAirConditionerControllerConfigurationData(string name)
                {
                    return new StructuredData(AirConditionerControllerConfigurationDataType)
                    {
                        // ReSharper disable once AssignNullToNotNullAttribute
                        {"Name", new PrimitiveData(UAOpcBinaryStandardDataTypes.CharArray, name)},
                        {"DeviceAddress", new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt32, (UInt32) 0)},
                        {"TemperatureSetpoint", new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, (double) 72)},
                        {"HumiditySetpoint", new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, (double) 60)}
                    };
                }

                var airConditionerControllersData = new List<GenericData>
                {
                    CreateAirConditionerControllerConfigurationData("AirConditioner_1"),
                    CreateAirConditionerControllerConfigurationData("AirConditioner_2"),
                    CreateAirConditionerControllerConfigurationData("AirConditioner_3"),
                    CreateAirConditionerControllerConfigurationData("AirConditioner_4"),
                    CreateAirConditionerControllerConfigurationData("AirConditioner_5"),
                    CreateAirConditionerControllerConfigurationData("AirConditioner_6"),
                    CreateAirConditionerControllerConfigurationData("AirConditioner_7"),
                    CreateAirConditionerControllerConfigurationData("AirConditioner_8"),
                    CreateAirConditionerControllerConfigurationData("AirConditioner_9"),
                    CreateAirConditionerControllerConfigurationData("AirConditioner_10")
                };

                StructuredData CreateFurnaceControllerConfigurationData(string name)
                {
                    return new StructuredData(ControllerConfigurationDataType)
                    {
                        // ReSharper disable once AssignNullToNotNullAttribute
                        {"Name", new PrimitiveData(UAOpcBinaryStandardDataTypes.CharArray, name)},
                        {"DeviceAddress", new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt32, (UInt32) 0)},
                        {"TemperatureSetpoint", new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, (double) 72)},
                    };
                }

                var furnaceControllersData = new List<GenericData>
                {
                    CreateFurnaceControllerConfigurationData("Furnace_1"),
                    CreateFurnaceControllerConfigurationData("Furnace_2"),
                    CreateFurnaceControllerConfigurationData("Furnace_3"),
                    CreateFurnaceControllerConfigurationData("Furnace_4"),
                    CreateFurnaceControllerConfigurationData("Furnace_5"),
                    CreateFurnaceControllerConfigurationData("Furnace_6"),
                    CreateFurnaceControllerConfigurationData("Furnace_7"),
                    CreateFurnaceControllerConfigurationData("Furnace_8"),
                    CreateFurnaceControllerConfigurationData("Furnace_9"),
                    CreateFurnaceControllerConfigurationData("Furnace_10")
                };

                // ReSharper disable CoVariantArrayConversion
                var result = new StructuredData(BAConfigurationsDataType)
                {
                    { "NoOfAirConditionerControllers", new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, 10) },
                    { "AirConditionerControllers", new SequenceData(AirConditionerControllerConfigurationArrayDataType, airConditionerControllersData) },
                    { "NoOfFurnaceControllers", new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, 10) },
                    { "FurnaceControllers", new SequenceData(ControllerConfigurationArrayDataType, furnaceControllersData) }
                };
                // ReSharper restore CoVariantArrayConversion
                return result;
            }
        }

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public StructuredDataType BAConfigurationsDataType
        {
            get
            {
                var result = new StructuredDataType(
                    "BAConfigurations",
                    "urn:UnifiedAutomation:CppDemoServer:BuildingAutomation:BAConfigurations")
                {
                    new DataField("NoOfAirConditionerControllers", UAOpcBinaryStandardDataTypes.Int32) { IsLength = true },
                    new DataField("AirConditionerControllers", AirConditionerControllerConfigurationArrayDataType) {IsOptional = true},
                    new DataField("NoOfFurnaceControllers", UAOpcBinaryStandardDataTypes.Int32) { IsLength = true },
                    new DataField("FurnaceControllers", ControllerConfigurationArrayDataType) {IsOptional = true}
                };
                return result;
            }
        }

        #endregion



        #region Static Private Properties

        [NotNull]
        static private SequenceDataType AirConditionerControllerConfigurationArrayDataType => new SequenceDataType(AirConditionerControllerConfigurationDataType);

        [NotNull]
        static private StructuredDataType AirConditionerControllerConfigurationDataType => new StructuredDataType(
                "AirConditionerControllerConfiguration",
                "urn:UnifiedAutomation:CppDemoServer:BuildingAutomation:AirConditionerControllerConfiguration",
                ControllerConfigurationDataType.DataFields)
            {
                new DataField("HumiditySetpoint", UAOpcBinaryStandardDataTypes.Double)
            };

        [NotNull]
        static private SequenceDataType ControllerConfigurationArrayDataType => new SequenceDataType(ControllerConfigurationDataType);

        [NotNull]
        static private StructuredDataType ControllerConfigurationDataType => new StructuredDataType(
                "ControllerConfiguration",
                "urn:UnifiedAutomation:CppDemoServer:BuildingAutomation:ControllerConfiguration")
            {
                new DataField("Name", UAOpcBinaryStandardDataTypes.CharArray),
                new DataField("DeviceAddress", UAOpcBinaryStandardDataTypes.UInt32),
                new DataField("TemperatureSetpoint", UAOpcBinaryStandardDataTypes.Double)
            };

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static private byte[] BAConfigurationsExtensionObjectBody
        {
            get
            {
                var bodyHexaString = "0A-00-00-00-10-00-00-00-41-69-72-43-6F-6E-64-69-74-69-6F-6E-65-72-5F-31-00-00-00-00-00-00-00-00-00-00-52-40-00-00-00-00-00-00-4E-40-10-00-00-00-41-69-72-43-6F-6E-64-69-74-69-6F-6E-65-72-5F-32-00-00-00-00-00-00-00-00-00-00-52-40-00-00-00-00-00-00-4E-40-10-00-00-00-41-69-72-43-6F-6E-64-69-74-69-6F-6E-65-72-5F-33-00-00-00-00-00-00-00-00-00-00-52-40-00-00-00-00-00-00-4E-40-10-00-00-00-41-69-72-43-6F-6E-64-69-74-69-6F-6E-65-72-5F-34-00-00-00-00-00-00-00-00-00-00-52-40-00-00-00-00-00-00-4E-40-10-00-00-00-41-69-72-43-6F-6E-64-69-74-69-6F-6E-65-72-5F-35-00-00-00-00-00-00-00-00-00-00-52-40-00-00-00-00-00-00-4E-40-10-00-00-00-41-69-72-43-6F-6E-64-69-74-69-6F-6E-65-72-5F-36-00-00-00-00-00-00-00-00-00-00-52-40-00-00-00-00-00-00-4E-40-10-00-00-00-41-69-72-43-6F-6E-64-69-74-69-6F-6E-65-72-5F-37-00-00-00-00-00-00-00-00-00-00-52-40-00-00-00-00-00-00-4E-40-10-00-00-00-41-69-72-43-6F-6E-64-69-74-69-6F-6E-65-72-5F-38-00-00-00-00-00-00-00-00-00-00-52-40-00-00-00-00-00-00-4E-40-10-00-00-00-41-69-72-43-6F-6E-64-69-74-69-6F-6E-65-72-5F-39-00-00-00-00-00-00-00-00-00-00-52-40-00-00-00-00-00-00-4E-40-11-00-00-00-41-69-72-43-6F-6E-64-69-74-69-6F-6E-65-72-5F-31-30-00-00-00-00-00-00-00-00-00-00-52-40-00-00-00-00-00-00-4E-40-0A-00-00-00-09-00-00-00-46-75-72-6E-61-63-65-5F-31-00-00-00-00-00-00-00-00-00-00-52-40-09-00-00-00-46-75-72-6E-61-63-65-5F-32-00-00-00-00-00-00-00-00-00-00-52-40-09-00-00-00-46-75-72-6E-61-63-65-5F-33-00-00-00-00-00-00-00-00-00-00-52-40-09-00-00-00-46-75-72-6E-61-63-65-5F-34-00-00-00-00-00-00-00-00-00-00-52-40-09-00-00-00-46-75-72-6E-61-63-65-5F-35-00-00-00-00-00-00-00-00-00-00-52-40-09-00-00-00-46-75-72-6E-61-63-65-5F-36-00-00-00-00-00-00-00-00-00-00-52-40-09-00-00-00-46-75-72-6E-61-63-65-5F-37-00-00-00-00-00-00-00-00-00-00-52-40-09-00-00-00-46-75-72-6E-61-63-65-5F-38-00-00-00-00-00-00-00-00-00-00-52-40-09-00-00-00-46-75-72-6E-61-63-65-5F-39-00-00-00-00-00-00-00-00-00-00-52-40-0A-00-00-00-46-75-72-6E-61-63-65-5F-31-30-00-00-00-00-00-00-00-00-00-00-52-40";
                byte[] extensionObjectBody = BitConverterUtilities.ArrayOfBytesFromString(bodyHexaString);
                return extensionObjectBody;
            }
        }

        [NotNull]
        static private byte[] TypeDictionaryByteArray
        {
            get
            {
                Type type = typeof(UAOpcBinaryDataTypeSystemTest);
                Stream stream = type.Assembly.GetManifestResourceStream(type, "BuildingAutomationTypeDictionary.xml");
                Debug.Assert(stream != null);
                byte[] dictionaryByteArray = stream.ReadToArrayOfBytes();
                return dictionaryByteArray;
            }
        }

        #endregion



        #region Static Private Methods

        static private void EncodeTest_BAConfigurations([NotNull] GenericData genericData)
        {
            EncodeTest(null, TypeDictionaryByteArray, "BAConfigurations", genericData, BAConfigurationsExtensionObjectBody);
        }

        #endregion



        #region Test Methods

        [TestMethod]
        public void DecodeTest_BAConfigurations()
        {
            DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "BAConfigurations",
                BAConfigurationsExtensionObjectBody,
                BAConfigurationsData);
        }

        [TestMethod]
        public void EncodeTest_BAConfigurations()
        {
            EncodeTest_BAConfigurations(BAConfigurationsData);
        }

        [TestMethod]
        public void EncodeTest_BAConfigurations_AutoCompleteLengthField()
        {
            var structuredData = (StructuredData)BAConfigurationsData;
            structuredData.FieldData.RemoveExisting("NoOfAirConditionerControllers");
            EncodeTest_BAConfigurations(structuredData);
        }

        // for profiling
        [TestMethod]
        public void RepeatedDecodeTest_BAConfigurations_100()
        {
            RepeatedDecodeTest(TypeDictionaryByteArray, "BAConfigurations",
                BAConfigurationsExtensionObjectBody,
                100);
        }

        // for profiling
        [TestMethod]
        public void RepeatedEncodeTest_BAConfigurations_100()
        {
            RepeatedEncodeTest(TypeDictionaryByteArray, "BAConfigurations",
                BAConfigurationsData,
                100);
        }

        #endregion
    }
}
