﻿
using System;
using System.Diagnostics;
using System.IO;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Opc.Ua.Schema;
using OpcLabs.BaseLib.Collections.Specialized;
using OpcLabs.BaseLib.DataTypeModel;
using OpcLabs.BaseLib.IO.Extensions;
using OpcLabs.BaseLib.IO.Utilities;
using OpcLabs.BaseLib.OperationModel.Generic;
using OpcLabs.BaseLib.Utilities;
using OpcLabs.EasyOpc.UA.DataTypeModel;
using OpcLabs.EasyOpc.UA.DataTypeModel.Implementation;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.UA.DataTypeModel.Internal
{
    [TestClass]
    public class StandardTypesTest : UAOpcBinaryDataTypeSystemTest
    {
        #region Static Private Properties

        [NotNull]
        static private byte[] TypeDictionaryByteArray
        {
            get
            {
                var uri = new Uri(UAStandardBinaryIntrinsicDataTypeDictionary.StandardTypesLocation);
                Stream stream = StreamUtilities.GetFromManagedResource(uri, reflectionOnlyLoad:false);
                Debug.Assert(stream != null);
                byte[] dictionaryByteArray = stream.ReadToArrayOfBytes();
                return dictionaryByteArray;
            }
        }

        #endregion



        #region Static Private Methods

        static private void EncodePrimitiveStandardDataTypeMismatchErrorTest(
            [NotNull] string dataTypeName,
            [NotNull] Type exceptType)
        {
            var testValues = new []
            {
                false,
                Byte.MinValue,
                Char.MinValue,
                DateTime.MinValue,
                Decimal.MinValue,
                Double.MinValue,
                Int16.MinValue,
                Int32.MinValue,
                Int64.MinValue,
                IntPtr.Zero,
                new object(),
                SByte.MinValue,
                Single.MinValue,
                String.Empty,
                UInt16.MinValue,
                UInt32.MinValue,
                UInt64.MinValue,
                UIntPtr.Zero
            };

            foreach (object testValue in testValues)
            {
                AssertUtilities.WriteLine();
                AssertUtilities.WriteLine($"testValue: {testValue}");
                Type testValueType = testValue.GetType();
                AssertUtilities.WriteLine($"testValueType: {testValueType}");

                if (testValueType == exceptType)
                    continue;

                EncodeErrorTest(
                    // ReSharper disable once PossibleNullReferenceException
                    dts => FindStandardBidtd(dts).AllowConvertiblePrimitiveDataValues = false, 
                    TypeDictionaryByteArray, 
                    dataTypeName, 
                    new PrimitiveData(testValue),
                    new[] { UADataTypeModelErrorCodes.OpcBinarySystemDictionaryEncoderFailure, UADataTypeModelErrorCodes.StandardDictionaryPrimitiveDataTypeMismatch});
            }
        }

        #endregion



        #region Test Methods

        // IMPROVE: put ths back after having way to configure the parameter AllowFileTimeOutOfDateTimeRange
        //[TestMethod]
        //public void DecodeErrorTest_StandardDictionaryRawDataFileTimeOutOfRange()
        //    => DecodeErrorTest(
        //        dts => dts.Parameters., 
        //        TypeDictionaryByteArray, "DateTime",
        //        new Byte[] { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x7F },
        //        UADataTypeModelErrorCodes.StandardDictionaryRawDataFileTimeOutOfRange);

        [TestMethod]
        public void DecodeErrorTest_OpcBinarySystemExtraDataPastEndOfStream()
            => DecodeErrorTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => dts.Parameters.AllowDataPastEndOfStream = false, 
                TypeDictionaryByteArray, "Int32",
                new Byte[] { 1, 0, 0, 0, 85 },
                UADataTypeModelErrorCodes.OpcBinarySystemExtraDataPastEndOfStream);



        [TestMethod]
        public void DecodeTest_Bit_False()
            => DecodeTest(null, TypeDictionaryByteArray, "Bit", 
                new Byte[] {0},
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false));

        [TestMethod]

        public void DecodeTest_Bit_True()
            => DecodeTest(null, TypeDictionaryByteArray, "Bit",
                new Byte[] { 1 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, true));

        [TestMethod]
        public void DecodeTest_Boolean_False()
            => DecodeTest(null, TypeDictionaryByteArray, "Boolean",
                new Byte[] { 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Boolean, false));

        [TestMethod]
        public void DecodeTest_Boolean_True_1()
            => DecodeTest(null, TypeDictionaryByteArray, "Boolean",
                new Byte[] { 1 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Boolean, true));

        [TestMethod]
        public void DecodeTest_Boolean_True_2()
            => DecodeTest(null, TypeDictionaryByteArray, "Boolean",
                new Byte[] { 2 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Boolean, true));

        [TestMethod]
        public void DecodeTest_Boolean_True_255()
            => DecodeTest(null, TypeDictionaryByteArray, "Boolean",
                new Byte[] { 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Boolean, true));

        [TestMethod]
        public void DecodeTest_Byte_0()
            => DecodeTest(null, TypeDictionaryByteArray, "Byte",
                new Byte[] { 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Byte, (Byte)0));

        [TestMethod]
        public void DecodeTest_Byte_1()
            => DecodeTest(null, TypeDictionaryByteArray, "Byte",
                new Byte[] { 1 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Byte, (Byte)1));

        [TestMethod]
        public void DecodeTest_Byte_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "Byte",
                new Byte[] { 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Byte, Byte.MaxValue));

        [TestMethod]
        public void DecodeTest_Char_0()
            => DecodeTest(null, TypeDictionaryByteArray, "Char",
                new Byte[] { 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Char, (Char)0));

        [TestMethod]
        public void DecodeTest_Char_1()
            => DecodeTest(null, TypeDictionaryByteArray, "Char",
                new Byte[] { 1 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Char, (Char)1));

        [TestMethod]
        public void DecodeTest_Char_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "Char",
                new Byte[] { 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Char, (Char)Byte.MaxValue));

        [TestMethod]
        public void DecodeTest_ByteString_Empty()
            => DecodeTest(null, TypeDictionaryByteArray, "ByteString",
                new Byte[] { 0, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.ByteString, new Byte[0]));

        [TestMethod]
        public void DecodeTest_ByteString_Three()
            => DecodeTest(null, TypeDictionaryByteArray, "ByteString",
                new Byte[] { 3, 0, 0, 0, 11, 22, 33 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.ByteString, new Byte[] {11, 22, 33}));

        [TestMethod]
        public void DecodeTest_CharArray_Empty()
            => DecodeTest(null, TypeDictionaryByteArray, "CharArray",
                new Byte[] { 0, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.CharArray, String.Empty));

        [TestMethod]
        public void DecodeTest_CharArray_Love()
            => DecodeTest(null, TypeDictionaryByteArray, "CharArray",
                new Byte[] { 4, 0, 0, 0, (Byte)'L', (Byte)'o', (Byte)'v', (Byte)'e' },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.CharArray, "Love"));

        [TestMethod]
        public void DecodeTest_DateTime_DateTime_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "DateTime",
                BitConverter.GetBytes(DateTime.MaxValue.ToFileTimeUtc()),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.DateTime, DateTimeUtilities.WeakSpecifyKind(DateTime.MaxValue, DateTimeKind.Utc)));

        [TestMethod]
        public void DecodeTest_DateTime_FileTime_0()
            => DecodeTest(null, TypeDictionaryByteArray, "DateTime",
                BitConverter.GetBytes((long)0),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.DateTime, DateTime.FromFileTimeUtc(0)));

        [TestMethod]
        public void DecodeTest_DateTime_FileTime_1()
            => DecodeTest(null, TypeDictionaryByteArray, "DateTime",
                BitConverter.GetBytes((long)1),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.DateTime, DateTime.FromFileTimeUtc(1)));

        [TestMethod]
        public void DecodeTest_DateTime_FileTime_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "DateTime",
                BitConverter.GetBytes(DateTimeUtilities.MaximumFileTime),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.DateTime, DateTime.FromFileTimeUtc(DateTimeUtilities.MaximumFileTime)));

        [TestMethod]
        public void DecodeTest_DateTime_FileTime_OutOfRange()
            => DecodeTest(null, TypeDictionaryByteArray, "DateTime",
                new Byte[] { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x7F },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.DateTime, DateTime.FromFileTimeUtc(DateTimeUtilities.MaximumFileTime)));

        [TestMethod]
        public void DecodeTest_Double_0()
            => DecodeTest(null, TypeDictionaryByteArray, "Double",
                BitConverter.GetBytes(0.0),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, 0.0));

        [TestMethod]
        public void DecodeTest_Double_1()
            => DecodeTest(null, TypeDictionaryByteArray, "Double",
                BitConverter.GetBytes(1.0),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, 1.0));

        [TestMethod]
        public void DecodeTest_Double_Epsilon()
            => DecodeTest(null, TypeDictionaryByteArray, "Double",
                BitConverter.GetBytes(Double.Epsilon),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, Double.Epsilon));

        [TestMethod]
        public void DecodeTest_Double_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "Double",
                BitConverter.GetBytes(Double.MaxValue),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, Double.MaxValue));

        [TestMethod]
        public void DecodeTest_Double_Minus1()
            => DecodeTest(null, TypeDictionaryByteArray, "Double",
                BitConverter.GetBytes(-1.0),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, -1.0));

        [TestMethod]
        public void DecodeTest_Double_MinValue()
            => DecodeTest(null, TypeDictionaryByteArray, "Double",
                BitConverter.GetBytes(Double.MinValue),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, Double.MinValue));

        [TestMethod]
        public void DecodeTest_Float_0()
            => DecodeTest(null, TypeDictionaryByteArray, "Float",
                BitConverter.GetBytes(0.0f),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Float, 0.0f));

        [TestMethod]
        public void DecodeTest_Float_1()
            => DecodeTest(null, TypeDictionaryByteArray, "Float",
                BitConverter.GetBytes(1.0f),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Float, 1.0f));

        [TestMethod]
        public void DecodeTest_Float_Epsilon()
            => DecodeTest(null, TypeDictionaryByteArray, "Float",
                BitConverter.GetBytes(Single.Epsilon),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Float, Single.Epsilon));

        [TestMethod]
        public void DecodeTest_Float_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "Float",
                BitConverter.GetBytes(Single.MaxValue),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Float, Single.MaxValue));

        [TestMethod]
        public void DecodeTest_Float_Minus1()
            => DecodeTest(null, TypeDictionaryByteArray, "Float",
                BitConverter.GetBytes(-1.0f),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Float, -1.0f));

        [TestMethod]
        public void DecodeTest_Float_MinValue()
            => DecodeTest(null, TypeDictionaryByteArray, "Float",
                BitConverter.GetBytes(Single.MinValue),
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Float, Single.MinValue));

        [TestMethod]
        public void DecodeTest_Guid_Empty()
            => DecodeTest(null, TypeDictionaryByteArray, "Guid",
                new Byte[16],
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Guid, Guid.Empty));

        [TestMethod]
        public void DecodeTest_Guid_Our()
        {
            Guid guid = Guid.Parse("2B0A4D0F-589B-4904-856D-7F913FDD2D5E");
            DecodeTest(null, TypeDictionaryByteArray, "Guid",
                new Byte[] { 0x0F, 0x4D, 0x0A, 0x2B, 0x9B, 0x58, 0x04, 0x49, 0x85, 0x6D, 0x7F, 0x91, 0x3F, 0xDD, 0x2D, 0x5E },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Guid, guid));
        }

        [TestMethod]
        public void DecodeTest_Int16_0()
            => DecodeTest(null, TypeDictionaryByteArray, "Int16",
                new Byte[] { 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int16, (Int16)0));

        [TestMethod]
        public void DecodeTest_Int16_1()
            => DecodeTest(null, TypeDictionaryByteArray, "Int16",
                new Byte[] { 1, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int16, (Int16)1));

        [TestMethod]
        public void DecodeTest_Int16_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "Int16",
                new Byte[] { 255, 127 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int16, Int16.MaxValue));

        [TestMethod]
        public void DecodeTest_Int16_Minus1()
            => DecodeTest(null, TypeDictionaryByteArray, "Int16",
                new Byte[] { 255, 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int16, (Int16)(-1)));

        [TestMethod]
        public void DecodeTest_Int16_MinValue()
            => DecodeTest(null, TypeDictionaryByteArray, "Int16",
                new Byte[] { 0, 128 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int16, Int16.MinValue));

        [TestMethod]
        public void DecodeTest_Int32_0()
            => DecodeTest(null, TypeDictionaryByteArray, "Int32",
                new Byte[] { 0, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, 0));

        [TestMethod]
        public void DecodeTest_Int32_1()
            => DecodeTest(null, TypeDictionaryByteArray, "Int32",
                new Byte[] { 1, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, 1));

        [TestMethod]
        public void DecodeTest_Int32_1_ExtraDataPastEndOfStream_Warning()
        {
            var expectedGenericData = new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, 1);
            DataType expectedDataType = expectedGenericData.DataType;
            Debug.Assert(expectedDataType != null);

            ValueResult<GenericData> decodedValueResult = DecodeTestCommon(null, TypeDictionaryByteArray, "Int32",
                new Byte[] {1, 0, 0, 0, 85});
            Assert.IsTrue(decodedValueResult.Succeeded, "Decoding has failed but should have succeeded.");
            GenericData decodedValue = decodedValueResult.Value;
            AssertUtilities.WriteLine($"decodedValue: {decodedValue:V}");
            Assert.IsNotNull(decodedValue, "Decoded value is a null reference.");
            Exception validationException = decodedValue.TryValidate();
            AssertUtilities.WriteLine($"validationException: {validationException}");
            Assert.IsNull(validationException, "Decoded value validation failed.");
            Assert.IsTrue(decodedValue.IsValid, "Decoded value is not valid.");
            Assert.AreEqual(expectedDataType.Kind, decodedValue.DataTypeKind, "Data type kinds differ.");
            AssertUtilities.WriteLine($"decodedValue.DataType: {decodedValue.DataType:V}");
            Assert.AreEqual(expectedDataType, decodedValue.DataType, "Data types differ.");
            Assert.AreEqual(expectedGenericData, decodedValue, "Decoded value not as expected.");

            ExceptionCollection diagnostics = decodedValueResult.Diagnostics;
            Assert.IsNotNull(diagnostics);
            Assert.AreEqual(1, diagnostics.Count);
            Exception warning = diagnostics[0];
            Assert.IsNotNull(warning);
            AssertUtilities.WriteLine(warning);
            string message = warning.Message;
            StringAssert.Contains(message, "data");
            StringAssert.Contains(message, "more");
            StringAssert.Contains(message, "stream");
        }

        [TestMethod]
        public void DecodeTest_Int32_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "Int32",
                new Byte[] { 255, 255, 255, 127 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, Int32.MaxValue));

        [TestMethod]
        public void DecodeTest_Int32_Minus1()
            => DecodeTest(null, TypeDictionaryByteArray, "Int32",
                new Byte[] { 255, 255, 255, 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, -1));

        [TestMethod]
        public void DecodeTest_Int32_MinValue()
            => DecodeTest(null, TypeDictionaryByteArray, "Int32",
                new Byte[] { 0, 0, 0, 128 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, Int32.MinValue));

        [TestMethod]
        public void DecodeTest_Int64_0()
            => DecodeTest(null, TypeDictionaryByteArray, "Int64",
                new Byte[] { 0, 0, 0, 0, 0, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int64, (Int64)0));

        [TestMethod]
        public void DecodeTest_Int64_1()
            => DecodeTest(null, TypeDictionaryByteArray, "Int64",
                new Byte[] { 1, 0, 0, 0, 0, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int64, (Int64)1));

        [TestMethod]
        public void DecodeTest_Int64_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "Int64",
                new Byte[] { 255, 255, 255, 255, 255, 255, 255, 127 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int64, Int64.MaxValue));

        [TestMethod]
        public void DecodeTest_Int64_Minus1()
            => DecodeTest(null, TypeDictionaryByteArray, "Int64",
                new Byte[] { 255, 255, 255, 255, 255, 255, 255, 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int64, (Int64)(-1)));

        [TestMethod]
        public void DecodeTest_Int64_MinValue()
            => DecodeTest(null, TypeDictionaryByteArray, "Int64",
                new Byte[] { 0, 0, 0, 0, 0, 0, 0, 128 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int64, Int64.MinValue));

        [TestMethod]
        public void DecodeTest_SByte_0()
            => DecodeTest(null, TypeDictionaryByteArray, "SByte",
                new Byte[] { 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, (Int16)0));

        [TestMethod]
        public void DecodeTest_SByte_0_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false, 
                TypeDictionaryByteArray, "SByte",
                new Byte[] { 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, (SByte)0));

        [TestMethod]
        public void DecodeTest_SByte_1()
            => DecodeTest(null, TypeDictionaryByteArray, "SByte",
                new Byte[] { 1 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, (Int16)1));

        [TestMethod]
        public void DecodeTest_SByte_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "SByte",
                new Byte[] { 1 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, (SByte)1));

        [TestMethod]
        public void DecodeTest_SByte_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "SByte",
                new Byte[] { 127 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, (Int16)SByte.MaxValue));

        [TestMethod]
        public void DecodeTest_SByte_MaxValue_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "SByte",
                new Byte[] { 127 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, SByte.MaxValue));

        [TestMethod]
        public void DecodeTest_SByte_Minus1()
            => DecodeTest(null, TypeDictionaryByteArray, "SByte",
                new Byte[] { 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, (Int16)(-1)));

        [TestMethod]
        public void DecodeTest_SByte_Minus1_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "SByte",
                new Byte[] { 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, (SByte)(-1)));

        [TestMethod]
        public void DecodeTest_SByte_MinValue()
            => DecodeTest(null, TypeDictionaryByteArray, "SByte",
                new Byte[] { 128 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, (Int16)SByte.MinValue));

        [TestMethod]
        public void DecodeTest_SByte_MinValue_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "SByte",
                new Byte[] { 128 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, SByte.MinValue));

        [TestMethod]
        public void DecodeTest_String_Empty()
            => DecodeTest(null, TypeDictionaryByteArray, "String",
                new Byte[] { 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.String, String.Empty));

        [TestMethod]
        public void DecodeTest_String_Love()
            => DecodeTest(null, TypeDictionaryByteArray, "String",
                new Byte[] { (Byte)'L', (Byte)'o', (Byte)'v', (Byte)'e', 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.String, "Love"));

        [TestMethod]
        public void DecodeTest_UInt16_0()
            => DecodeTest(null, TypeDictionaryByteArray, "UInt16",
                new Byte[] { 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt16, /*(Int32)*/0));

        [TestMethod]
        public void DecodeTest_UInt16_0_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "UInt16",
                new Byte[] { 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt16, (UInt16)0));

        [TestMethod]
        public void DecodeTest_UInt16_1()
            => DecodeTest(null, TypeDictionaryByteArray, "UInt16",
                new Byte[] { 1, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt16, /*(Int32)*/1));

        [TestMethod]
        public void DecodeTest_UInt16_1_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "UInt16",
                new Byte[] { 1, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt16, (UInt16)1));

        [TestMethod]
        public void DecodeTest_UInt16_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "UInt16",
                new Byte[] { 255, 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt16, (Int32)UInt16.MaxValue));

        [TestMethod]
        public void DecodeTest_UInt16_MaxValue_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "UInt16",
                new Byte[] { 255, 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt16, UInt16.MaxValue));

        [TestMethod]
        public void DecodeTest_UInt32_0()
            => DecodeTest(null, TypeDictionaryByteArray, "UInt32",
                new Byte[] { 0, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt32, (Int64)0));

        [TestMethod]
        public void DecodeTest_UInt32_0_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "UInt32",
                new Byte[] { 0, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt32, (UInt32)0));

        [TestMethod]
        public void DecodeTest_UInt32_1()
            => DecodeTest(null, TypeDictionaryByteArray, "UInt32",
                new Byte[] { 1, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt32, (Int64)1));

        [TestMethod]
        public void DecodeTest_UInt32_1_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "UInt32",
                new Byte[] { 1, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt32, (UInt32)1));

        [TestMethod]
        public void DecodeTest_UInt32_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "UInt32",
                new Byte[] { 255, 255, 255, 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt32, (Int64)UInt32.MaxValue));

        [TestMethod]
        public void DecodeTest_UInt32_MaxValue_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "UInt32",
                new Byte[] { 255, 255, 255, 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt32, UInt32.MaxValue));

        [TestMethod]
        public void DecodeTest_UInt64_0()
            => DecodeTest(null, TypeDictionaryByteArray, "UInt64",
                new Byte[] { 0, 0, 0, 0, 0, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt64, (Decimal)0));

        [TestMethod]
        public void DecodeTest_UInt64_0_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "UInt64",
                new Byte[] { 0, 0, 0, 0, 0, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt64, (UInt64)0));

        [TestMethod]
        public void DecodeTest_UInt64_1()
            => DecodeTest(null, TypeDictionaryByteArray, "UInt64",
                new Byte[] { 1, 0, 0, 0, 0, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt64, (Decimal)1));

        [TestMethod]
        public void DecodeTest_UInt64_1_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "UInt64",
                new Byte[] { 1, 0, 0, 0, 0, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt64, (UInt64)1));

        [TestMethod]
        public void DecodeTest_UInt64_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "UInt64",
                new Byte[] { 255, 255, 255, 255, 255, 255, 255, 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt64, (Decimal)UInt64.MaxValue));

        [TestMethod]
        public void DecodeTest_UInt64_MaxValue_NoCls()
            => DecodeTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).ReturnClsCompliantPrimitiveTypes = false,
                TypeDictionaryByteArray, "UInt64",
                new Byte[] { 255, 255, 255, 255, 255, 255, 255, 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt64, UInt64.MaxValue));

        [TestMethod]
        public void DecodeTest_WideChar_0()
            => DecodeTest(null, TypeDictionaryByteArray, "WideChar",
                new Byte[] { 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideChar, (Char)0));

        [TestMethod]
        public void DecodeTest_WideChar_1()
            => DecodeTest(null, TypeDictionaryByteArray, "WideChar",
                new Byte[] { 1, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideChar, (Char)1));

        [TestMethod]
        public void DecodeTest_WideChar_MaxValue()
            => DecodeTest(null, TypeDictionaryByteArray, "WideChar",
                new Byte[] { 255, 255 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideChar, Char.MaxValue));

        [TestMethod]
        public void DecodeTest_WideCharArray_Empty()
            => DecodeTest(null, TypeDictionaryByteArray, "WideCharArray",
                new Byte[] { 0, 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideCharArray, String.Empty));

        [TestMethod]
        public void DecodeTest_WideCharArray_Euro()
            => DecodeTest(null, TypeDictionaryByteArray, "WideCharArray",
                new Byte[] { 1, 0, 0, 0, 0xAC, 0x20 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideCharArray, "\x20AC"));

        [TestMethod]
        public void DecodeTest_WideCharArray_Love()
            => DecodeTest(null, TypeDictionaryByteArray, "WideCharArray",
                new Byte[] { 4, 0, 0, 0, (Byte)'L', 0, (Byte)'o', 0, (Byte)'v', 0, (Byte)'e', 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideCharArray, "Love"));

        [TestMethod]
        public void DecodeTest_WideString_Empty()
            => DecodeTest(null, TypeDictionaryByteArray, "WideString",
                new Byte[] { 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideString, String.Empty));

        [TestMethod]
        public void DecodeTest_WideString_Euro()
            => DecodeTest(null, TypeDictionaryByteArray, "WideString",
                new Byte[] { 0xAC, 0x20, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideString, "\x20AC"));

        [TestMethod]
        public void DecodeTest_WideString_Love()
            => DecodeTest(null, TypeDictionaryByteArray, "WideString",
                new Byte[] { (Byte)'L', 0, (Byte)'o', 0, (Byte)'v', 0, (Byte)'e', 0, 0, 0 },
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideString, "Love"));



        [TestMethod]
        public void EncodeErrorTest_BinaryEncoderPrimitiveStandardDataDateTimeOutOfRange()
            => EncodeErrorTest(null, TypeDictionaryByteArray, "DateTime",
                new PrimitiveData(DateTime.MinValue),
                new[] { UADataTypeModelErrorCodes.OpcBinarySystemDictionaryEncoderFailure, UADataTypeModelErrorCodes.StandardDictionaryPrimitiveDataDateTimeOutOfRange});



        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_Bit() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("Bit", typeof(Boolean));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_Boolean() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("Boolean", typeof(Boolean));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_Byte() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("Byte", typeof(Byte));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_ByteString() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("ByteString", typeof(String));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_Char() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("Char", typeof(Char));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_CharArray() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("CharArray", typeof(String));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_DateTime() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("DateTime", typeof(DateTime));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_Double() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("Double", typeof(Double));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_Float() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("Float", typeof(Single));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_Guid() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("Guid", typeof(Guid));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_Int16() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("Int16", typeof(Int16));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_Int32() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("Int32", typeof(Int32));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_Int64() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("Int64", typeof(Int64));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_SByte() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("SByte", typeof(SByte));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_String() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("String", typeof(String));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_UInt16() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("UInt16", typeof(UInt16));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_UInt32() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("UInt32", typeof(UInt32));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_UInt64() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("UInt64", typeof(UInt64));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_WideChar() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("WideChar", typeof(Char));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_WideCharArray() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("WideCharArray", typeof(String));

        [TestMethod]
        public void EncodePrimitiveStandardDataTypeMismatchErrorTests_WideString() =>
            EncodePrimitiveStandardDataTypeMismatchErrorTest("WideString", typeof(String));



        [TestMethod]
        public void EncodeTest_Bit_False()
            => EncodeTest(null, TypeDictionaryByteArray, "Bit",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, false),
                new Byte[] {0});

        [TestMethod]
        public void EncodeTest_Bit_True()
            => EncodeTest(null, TypeDictionaryByteArray, "Bit",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Bit, true),
                new Byte[] {1});

        [TestMethod]
        public void EncodeTest_Boolean_False()
            => EncodeTest(null, TypeDictionaryByteArray, "Boolean",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Boolean, false),
                new Byte[] { 0 });

        [TestMethod]
        public void EncodeTest_Boolean_True()
            => EncodeTest(null, TypeDictionaryByteArray, "Boolean",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Boolean, true),
                new Byte[] { 255 });

        [TestMethod]
        public void EncodeTest_Byte_0()
            => EncodeTest(null, TypeDictionaryByteArray, "Byte",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Byte, (Byte)0),
                new Byte[] { 0 });

        [TestMethod]
        public void EncodeTest_Byte_1()
            => EncodeTest(null, TypeDictionaryByteArray, "Byte",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Byte, (Byte)1),
                new Byte[] { 1 });

        [TestMethod]
        public void EncodeTest_Byte_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "Byte",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Byte, Byte.MaxValue),
                new Byte[] { 255 });

        [TestMethod]
        public void EncodeTest_ByteString_Empty()
            => EncodeTest(null, TypeDictionaryByteArray, "ByteString",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.ByteString, new Byte[0]),
                new Byte[] { 0, 0, 0, 0 });

        [TestMethod]
        public void EncodeTest_ByteString_Three()
            => EncodeTest(null, TypeDictionaryByteArray, "ByteString",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.ByteString, new Byte[] {11, 22, 33}),
                new Byte[] { 3, 0, 0, 0, 11, 22, 33 });

        [TestMethod]
        public void EncodeTest_Char_0()
            => EncodeTest(null, TypeDictionaryByteArray, "Char",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Char, (Char)0),
                new Byte[] { 0 });

        [TestMethod]
        public void EncodeTest_Char_1()
            => EncodeTest(null, TypeDictionaryByteArray, "Char",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Char, (Char)1),
                new Byte[] { 1 });

        [TestMethod]
        public void EncodeTest_Char_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "Char",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Char, (Char)Byte.MaxValue),
                new Byte[] { 255 });

        [TestMethod]
        public void EncodeTest_CharArray_Empty()
            => EncodeTest(null, TypeDictionaryByteArray, "CharArray",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.CharArray, String.Empty),
                new Byte[] { 0, 0, 0, 0 });

        [TestMethod]
        public void EncodeTest_CharArray_Love()
            => EncodeTest(null, TypeDictionaryByteArray, "CharArray",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.CharArray, "Love"),
                new Byte[] { 4, 0, 0, 0, (Byte)'L', (Byte)'o', (Byte)'v', (Byte)'e' });

        [TestMethod]
        public void EncodeTest_DateTime_FileTime_0()
            => EncodeTest(null, TypeDictionaryByteArray, "DateTime",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.DateTime, DateTime.FromFileTimeUtc(0)),
                BitConverter.GetBytes((long)0));

        [TestMethod]
        public void EncodeTest_DateTime_FileTime_1()
            => EncodeTest(null, TypeDictionaryByteArray, "DateTime",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.DateTime, DateTime.FromFileTimeUtc(1)),
                BitConverter.GetBytes((long)1));

        [TestMethod]
        public void EncodeTest_DateTime_FileTime_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "DateTime",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.DateTime, DateTime.FromFileTimeUtc(DateTimeUtilities.MaximumFileTime)),
                BitConverter.GetBytes(DateTimeUtilities.MaximumFileTime));

        [TestMethod]
        public void EncodeTest_DateTime_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "DateTime",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.DateTime, DateTime.MaxValue),
                BitConverter.GetBytes(DateTime.MaxValue.ToFileTimeUtc()));

        [TestMethod]
        public void EncodeTest_Double_0()
            => EncodeTest(null, TypeDictionaryByteArray, "Double",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, 0.0),
                BitConverter.GetBytes(0.0));

        [TestMethod]
        public void EncodeTest_Double_1()
            => EncodeTest(null, TypeDictionaryByteArray, "Double",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, 1.0),
                BitConverter.GetBytes(1.0));

        [TestMethod]
        public void EncodeTest_Double_Epsilon()
            => EncodeTest(null, TypeDictionaryByteArray, "Double",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, Double.Epsilon),
                BitConverter.GetBytes(Double.Epsilon));

        [TestMethod]
        public void EncodeTest_Double_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "Double",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, Double.MaxValue),
                BitConverter.GetBytes(Double.MaxValue));

        [TestMethod]
        public void EncodeTest_Double_Minus1()
            => EncodeTest(null, TypeDictionaryByteArray, "Double",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, -1.0),
                BitConverter.GetBytes(-1.0));

        [TestMethod]
        public void EncodeTest_Double_MinValue()
            => EncodeTest(null, TypeDictionaryByteArray, "Double",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Double, Double.MinValue),
                BitConverter.GetBytes(Double.MinValue));

        [TestMethod]
        public void EncodeTest_Float_0()
            => EncodeTest(null, TypeDictionaryByteArray, "Float",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Float, 0.0f),
                BitConverter.GetBytes(0.0f));

        [TestMethod]
        public void EncodeTest_Float_1()
            => EncodeTest(null, TypeDictionaryByteArray, "Float",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Float, 1.0f),
                BitConverter.GetBytes(1.0f));

        [TestMethod]
        public void EncodeTest_Float_Epsilon()
            => EncodeTest(null, TypeDictionaryByteArray, "Float",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Float, Single.Epsilon),
                BitConverter.GetBytes(Single.Epsilon));

        [TestMethod]
        public void EncodeTest_Float_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "Float",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Float, Single.MaxValue),
                BitConverter.GetBytes(Single.MaxValue));

        [TestMethod]
        public void EncodeTest_Float_Minus1()
            => EncodeTest(null, TypeDictionaryByteArray, "Float",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Float, -1.0f),
                BitConverter.GetBytes(-1.0f));

        [TestMethod]
        public void EncodeTest_Float_MinValue()
            => EncodeTest(null, TypeDictionaryByteArray, "Float",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Float, Single.MinValue),
                BitConverter.GetBytes(Single.MinValue));

        [TestMethod]
        public void EncodeTest_Guid_Empty()
            => EncodeTest(null, TypeDictionaryByteArray, "Guid",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Guid, Guid.Empty),
                new Byte[16]);

        [TestMethod]
        public void EncodeTest_Guid_Our()
        {
            Guid guid = Guid.Parse("2B0A4D0F-589B-4904-856D-7F913FDD2D5E");
            EncodeTest(null, TypeDictionaryByteArray, "Guid",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Guid, guid),
                new Byte[] { 0x0F , 0x4D , 0x0A , 0x2B , 0x9B , 0x58 , 0x04 , 0x49 , 0x85 , 0x6D , 0x7F , 0x91 , 0x3F , 0xDD , 0x2D , 0x5E });
        }

        [TestMethod]
        public void EncodeTest_Int16_0()
            => EncodeTest(null, TypeDictionaryByteArray, "Int16",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int16, (Int16)0),
                new Byte[] { 0, 0 });

        [TestMethod]
        public void EncodeTest_Int16_1()
            => EncodeTest(null, TypeDictionaryByteArray, "Int16",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int16, (Int16)1),
                new Byte[] { 1, 0 });

        [TestMethod]
        public void EncodeTest_Int16_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "Int16",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int16, Int16.MaxValue),
                new Byte[] { 255, 127 });

        [TestMethod]
        public void EncodeTest_Int16_Minus1()
            => EncodeTest(null, TypeDictionaryByteArray, "Int16",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int16, (Int16)(-1)),
                new Byte[] { 255 , 255});

        [TestMethod]
        public void EncodeTest_Int16_MinValue()
            => EncodeTest(null, TypeDictionaryByteArray, "Int16",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int16, Int16.MinValue),
                new Byte[] { 0, 128 });

        [TestMethod]
        public void EncodeTest_Int32_0()
            => EncodeTest(null, TypeDictionaryByteArray, "Int32",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, 0),
                new Byte[] { 0, 0, 0, 0 });

        [TestMethod]
        public void EncodeTest_Int32_1()
            => EncodeTest(null, TypeDictionaryByteArray, "Int32",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, 1),
                new Byte[] { 1, 0, 0, 0 });

        [TestMethod]
        public void EncodeTest_Int32_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "Int32",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, Int32.MaxValue),
                new Byte[] { 255, 255, 255, 127 });

        [TestMethod]
        public void EncodeTest_Int32_Minus1()
            => EncodeTest(null, TypeDictionaryByteArray, "Int32",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, -1),
                new Byte[] { 255, 255, 255, 255 });

        [TestMethod]
        public void EncodeTest_Int32_MinValue()
            => EncodeTest(null, TypeDictionaryByteArray, "Int32",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int32, Int32.MinValue),
                new Byte[] { 0, 0, 0, 128 });

        [TestMethod]
        public void EncodeTest_Int64_0()
            => EncodeTest(null, TypeDictionaryByteArray, "Int64",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int64, (Int64)0),
                new Byte[] { 0, 0, 0, 0, 0, 0, 0, 0 });

        [TestMethod]
        public void EncodeTest_Int64_1()
            => EncodeTest(null, TypeDictionaryByteArray, "Int64",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int64, (Int64)1),
                new Byte[] { 1, 0, 0, 0, 0, 0, 0, 0 });

        [TestMethod]
        public void EncodeTest_Int64_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "Int64",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int64, Int64.MaxValue),
                new Byte[] { 255, 255, 255, 255, 255, 255, 255, 127 });

        [TestMethod]
        public void EncodeTest_Int64_Minus1()
            => EncodeTest(null, TypeDictionaryByteArray, "Int64",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int64, (Int64)(-1)),
                new Byte[] { 255, 255, 255, 255, 255, 255, 255, 255 });

        [TestMethod]
        public void EncodeTest_Int64_MinValue()
            => EncodeTest(null, TypeDictionaryByteArray, "Int64",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.Int64, Int64.MinValue),
                new Byte[] { 0, 0, 0, 0, 0, 0, 0, 128 });

        [TestMethod]
        public void EncodeTest_SByte_0()
            => EncodeTest(null, TypeDictionaryByteArray, "SByte",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, (SByte)0),
                new Byte[] { 0 });

        [TestMethod]
        public void EncodeTest_SByte_1()
            => EncodeTest(null, TypeDictionaryByteArray, "SByte",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, (SByte)1),
                new Byte[] { 1 });

        [TestMethod]
        public void EncodeTest_SByte_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "SByte",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, SByte.MaxValue),
                new Byte[] { 127 });

        [TestMethod]
        public void EncodeTest_SByte_Minus1()
            => EncodeTest(null, TypeDictionaryByteArray, "SByte",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, (SByte)(-1)),
                new Byte[] { 255 });

        [TestMethod]
        public void EncodeTest_SByte_MinValue()
            => EncodeTest(null, TypeDictionaryByteArray, "SByte",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.SByte, SByte.MinValue),
                new Byte[] { 128 });

        [TestMethod]
        public void EncodeTest_String_Empty()
            => EncodeTest(null, TypeDictionaryByteArray, "String",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.String, String.Empty),
                new Byte[] { 0 });

        [TestMethod]
        public void EncodeTest_String_Love()
            => EncodeTest(null, TypeDictionaryByteArray, "String",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.String, "Love"),
                new Byte[] { (Byte)'L', (Byte)'o', (Byte)'v', (Byte)'e', 0 });

        [TestMethod]
        public void EncodeTest_UInt16_0()
            => EncodeTest(null, TypeDictionaryByteArray, "UInt16",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt16, (UInt16)0),
                new Byte[] { 0, 0 });

        [TestMethod]
        public void EncodeTest_UInt16_1()
            => EncodeTest(null, TypeDictionaryByteArray, "UInt16",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt16, (UInt16)1),
                new Byte[] { 1, 0 });

        [TestMethod]
        public void EncodeTest_UInt16_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "UInt16",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt16, UInt16.MaxValue),
                new Byte[] { 255, 255 });

        [TestMethod]
        public void EncodeTest_UInt32_0()
            => EncodeTest(null, TypeDictionaryByteArray, "UInt32",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt32, (UInt32)0),
                new Byte[] { 0, 0, 0, 0 });

        [TestMethod]
        public void EncodeTest_UInt32_1()
            => EncodeTest(null, TypeDictionaryByteArray, "UInt32",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt32, (UInt32)1),
                new Byte[] { 1, 0, 0, 0 });

        [TestMethod]
        public void EncodeTest_UInt32_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "UInt32",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt32, UInt32.MaxValue),
                new Byte[] { 255, 255, 255, 255 });

        [TestMethod]
        public void EncodeTest_UInt64_0()
            => EncodeTest(null, TypeDictionaryByteArray, "UInt64",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt64, (UInt64)0),
                new Byte[] { 0, 0, 0, 0, 0, 0, 0, 0 });

        [TestMethod]
        public void EncodeTest_UInt64_1()
            => EncodeTest(null, TypeDictionaryByteArray, "UInt64",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt64, (UInt64)1),
                new Byte[] { 1, 0, 0, 0, 0, 0, 0, 0 });

        [TestMethod]
        public void EncodeTest_UInt64_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "UInt64",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.UInt64, UInt64.MaxValue),
                new Byte[] { 255, 255, 255, 255, 255, 255, 255, 255 });

        [TestMethod]
        public void EncodeTest_WideChar_0()
            => EncodeTest(null, TypeDictionaryByteArray, "WideChar",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideChar, (Char)0),
                new Byte[] { 0, 0 });

        [TestMethod]
        public void EncodeTest_WideChar_1()
            => EncodeTest(null, TypeDictionaryByteArray, "WideChar",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideChar, (Char)1),
                new Byte[] { 1, 0 });

        [TestMethod]
        public void EncodeTest_WideChar_MaxValue()
            => EncodeTest(null, TypeDictionaryByteArray, "WideChar",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideChar, Char.MaxValue),
                new Byte[] { 255, 255 });

        [TestMethod]
        public void EncodeTest_WideCharArray_Empty()
            => EncodeTest(null, TypeDictionaryByteArray, "WideCharArray",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideCharArray, String.Empty),
                new Byte[] { 0, 0, 0, 0 });

        [TestMethod]
        public void EncodeTest_WideCharArray_Euro()
            => EncodeTest(null, TypeDictionaryByteArray, "WideCharArray",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideCharArray, "\x20AC"),
                new Byte[] { 1, 0, 0, 0, 0xAC, 0x20 });

        [TestMethod]
        public void EncodeTest_WideCharArray_Love()
            => EncodeTest(null, TypeDictionaryByteArray, "WideCharArray",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideCharArray, "Love"),
                new Byte[] { 4, 0, 0, 0, (Byte)'L', 0, (Byte)'o', 0, (Byte)'v', 0, (Byte)'e', 0 });

        [TestMethod]
        public void EncodeTest_WideString_Empty()
            => EncodeTest(null, TypeDictionaryByteArray, "WideString",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideString, String.Empty),
                new Byte[] { 0, 0 });

        [TestMethod]
        public void EncodeTest_WideString_Euro()
            => EncodeTest(null, TypeDictionaryByteArray, "WideString",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideString, "\x20AC"),
                new Byte[] { 0xAC, 0x20, 0, 0 });

        [TestMethod]
        public void EncodeTest_WideString_Love()
            => EncodeTest(null, TypeDictionaryByteArray, "WideString",
                new PrimitiveData(UAOpcBinaryStandardDataTypes.WideString, "Love"),
                new Byte[] { (Byte)'L', 0, (Byte)'o', 0, (Byte)'v', 0, (Byte)'e', 0, 0, 0 });

        #endregion
    }
}
